#!/usr/bin/env python3
# coding: utf-8

from .version import __version__ as version

__author__ = "Axel Cournac, Axel Breuer, Lyam Baudry, Rémi Montagne, Cyril Matthey-Doret, Nadège Guielguimoni"
__copyright__ = "Copyright © 2017-2018, Institut Pasteur, Paris, France"
__credits__ = ["Hackathon ENGIE/Institut Pasteur teams"]
__license__ = "GPLv3"
__maintainer__ = "Axel Cournac, Lyam Baudry, Cyril Matthey-Doret, Nadège Guielguimoni"
__email__ = "axel.cournac@pasteur.fr"
__status__ = "Pre-Alpha"
__version__ = version
